#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <windows.h>
#include <mapi.h>
#include "arg1/arg1.h"

lpMapiRecipDesc Recips = NULL;
lpMapiFileDesc  Files  = NULL;
char*           Text   = NULL;
FLAGS flSendMail = MAPI_DIALOG;

const MapiRecipDesc defaultRecip = {
  0,
  MAPI_TO, // ulRecipClass Orig/To/Cc/Bcc
  NULL,    // lpszName     name
  NULL,    // lpszAddress  email
  0,
  NULL
};

const MapiFileDesc defaultFile = {
  0,
  0,
  -1,   // nPosition    in text
  NULL, // lpszPathName full path
  NULL,
  NULL,
};

MapiMessage message = {
  0,
  NULL, // lpszSubject  subject
  NULL, // lpszNoteText message
  NULL,
  NULL,
  NULL,
  0,
  NULL, // lpOriginator originator
  0,    // nRecipCount  recipient count
  NULL, // lpRecips     recipient array
  0,    // nFileCount   attachment count
  NULL  // lpFiles      attachment array
};

enum eOpt {
  eOpt_To,
  eOpt_Cc,
  eOpt_Bcc,
  eOpt_Subject,
  eOpt_Msg,
  eOpt_TextfileName,
  eOpt_Attach,
  eOpt_Quiet,
  eOpt_NOF,
};

const char* Options[eOpt_NOF] = {
  "-r",
  "-c",
  "-b",
  "-s",
  "-m",
  "-t",
  "-f",
  "-q",
};

static enum eOpt GetOpt(const char* Opt);
static void ReadTextfile(const char* TextfileName);
static int  SendMapi(void);
static void Exit(char ExitCode);

/*****************************************************************************
* main
*****************************************************************************/
int main(int argc, const char* argv[]) {
  /* Parse calling params                                                   */
  int i;
  int iRecip = 0;
  int nRecip = 0;
  int iFile  = 0;
  int nFile  = 0;
  const char* TextfileName = NULL;

  arg1Eval(argv[1]);

  for (i = 1; i < argc; ++i) {
    enum eOpt Opt = GetOpt(argv[i]);
    switch (Opt) {
      case eOpt_To:
      case eOpt_Cc:
      case eOpt_Bcc:
        if (i < argc - 1) nRecip++;
        break;
      case eOpt_TextfileName:
        if (i < argc - 1) TextfileName = argv[i + 1];
        break;
      case eOpt_Attach:
        if (i < argc - 1) nFile++;
        break;
      case eOpt_Quiet:
        flSendMail = 0;
        break;
      default:
        break;
    }
  }

  /* Allocate memory for and init recip and file objects                    */
  if (nRecip) {
    Recips = malloc(sizeof(MapiRecipDesc) * nRecip);
    if (Recips == NULL) {
      fprintf(stderr, "Memory allocation for recipients failed.\n");
      Exit(1);
    }
    else {
      for (i = 0; i < nRecip; ++i) memcpy(Recips + i,
                                          &defaultRecip,
                                          sizeof(defaultRecip));
      message.nRecipCount = nRecip;
      message.lpRecips = Recips;
    }
  }

  if (nFile) {
    Files = malloc(sizeof(MapiFileDesc) * nFile);
    if (Files == NULL) {
      fprintf(stderr, "Memory allocation for attachments failed.\n");
      Exit(1);
    }
    else {
      for (i = 0; i < nFile; ++i) memcpy(Files + i,
                                         &defaultFile,
                                         sizeof(defaultFile));
      message.nFileCount = nFile;
      message.lpFiles = Files;
    }
  }

  ReadTextfile(TextfileName);

  /* Fill up recip, file and message objects                                */
  for (i = 1; i < argc - 1; ++i) {
    enum eOpt Opt = GetOpt(argv[i]);
    switch (Opt) {
      case eOpt_To:
      case eOpt_Cc:
      case eOpt_Bcc:
        if (strchr(argv[i + 1], '@')) Recips[iRecip].lpszAddress = argv[i + 1];
        /* Outlook wants a name */    Recips[iRecip].lpszName    = argv[i + 1];
        Recips[iRecip].ulRecipClass = MAPI_TO + Opt - eOpt_To;
        iRecip++;
        break;
      case eOpt_Subject:
        message.lpszSubject = argv[i + 1];
        break;
      case eOpt_Msg:
        message.lpszNoteText = argv[i + 1];
        break;
      case eOpt_TextfileName:
        message.lpszNoteText = Text;
        break;
      case eOpt_Attach:
        Files[iFile].lpszPathName = argv[i + 1];
        iFile++;
        break;
      default:
        break;
    }

  }

  /* Send mail                                                              */
  Exit(SendMapi());
}


/*****************************************************************************
* Get option index
*****************************************************************************/
static enum eOpt GetOpt(const char* Opt) {
  enum eOpt i;
  for (i = 0; i < eOpt_NOF; i++) if (strcmp(Opt, Options[i]) == 0) break;
  return i;
}

/*****************************************************************************
* Read message body from text file to buffer
*****************************************************************************/
static void ReadTextfile(const char* TextfileName) {
  if (TextfileName) {
    FILE* Textfile = fopen(TextfileName, "r");
    if (Textfile == NULL) {
      fprintf(stderr, "Textfile for message body could not be opened.\n");
      Exit(1);
    }
    else {
      int Size;
      fseek(Textfile, 0, SEEK_END);
      Size = ftell(Textfile);

      Text = malloc(Size);
      if (Text == NULL) {
        fprintf(stderr, "Text buffer could not be allocated.\n");
        fclose(Textfile);
        Exit(1);
      }
      else {
        fseek(Textfile, 0, SEEK_SET);
        fread(Text, 1, Size, Textfile);
        Text[Size - 1] = 0;
        fclose(Textfile);
      }
    }
  }
}

/*****************************************************************************
* Send email via MAPI
*****************************************************************************/
static int  SendMapi(void) {
  int i;
  int Error = 0;

  /* Error handling declarations                                            */
  static const char* MapiErrors[] = {
    #include "MapiError.h"
    "UNKNOWN_ERROR\n",
  };
  #define N_MAPI_ERRORS (sizeof(MapiErrors) / sizeof(char*))

  static const char* ResultText[] = {
    "Logon    error : ",
    "SendMail error : ",
    "Logoff   error : ",
  };
  ULONG Result[3];

  /* The actual Windows stuff                                               */
  LHANDLE lhSession;
  HANDLE hMAPILib = LoadLibrary("MAPI32.DLL");

  LPMAPILOGON    lpfnMAPILogon    =
      (LPMAPILOGON)   GetProcAddress(hMAPILib, "MAPILogon");
  LPMAPISENDMAIL lpfnMAPISendMail =
      (LPMAPISENDMAIL)GetProcAddress(hMAPILib, "MAPISendMail");
  LPMAPILOGOFF   lpfnMAPILogoff   =
      (LPMAPILOGOFF)  GetProcAddress(hMAPILib, "MAPILogoff");

  Result[0] = (*lpfnMAPILogon)   (0, NULL, NULL, 1, 0, &lhSession);
  Result[1] = (*lpfnMAPISendMail)(0, 0, &message, flSendMail, 0);
  Result[2] = (*lpfnMAPILogoff)  (lhSession, 0, 0, 0);

  FreeLibrary(hMAPILib);

  /* Error info to stderr                                                   */
  for (i = 0; i < 3; ++i)
  if  (Result[i]) {
    Error = 1;
    fprintf(stderr, ResultText[i]);
    fprintf(stderr, "%10d ", Result[i]);

    if (Result[i] >= N_MAPI_ERRORS) Result[i] = N_MAPI_ERRORS - 1;
    fprintf(stderr, MapiErrors[Result[i]]);
  }
  return Error;
}

/*****************************************************************************
* Safe exit freeing memory
*****************************************************************************/
static void Exit(char ExitCode) {
  /* Destroy recip and file objects                                         */
  if (Recips) free(Recips);
  if (Files)  free(Files);
  if (Text)   free(Text);
  exit(ExitCode);
}
